import inspect
import types
import warnings
from importlib import import_module

from django.conf import settings as django_settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractBaseUser
from django.core import signing
from django.utils.translation import gettext as _
from graphene_django.utils import camelize

from .exceptions import TokenScopeError

warnings.simplefilter("once")


def get_token(user, action, **kwargs):
    username = user.get_username()
    if hasattr(username, "pk"):
        username = username.pk
    payload = {user.USERNAME_FIELD: username, "action": action}
    if kwargs:
        payload.update(**kwargs)
    token = signing.dumps(payload)
    return token


def get_token_payload(token, action, exp=None):
    payload = signing.loads(token, max_age=exp)
    _action = payload.pop("action")
    if _action != action:
        raise TokenScopeError
    return payload


def get_token_paylod(token, action, exp=None):
    warnings.warn(
        "get_token_paylod is deprecated, use get_token_payload instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return get_token_payload(token, action, exp)


def using_refresh_tokens():
    return (
        hasattr(django_settings, "GRAPHQL_JWT")
        and django_settings.GRAPHQL_JWT.get("JWT_LONG_RUNNING_REFRESH_TOKEN", False)
        and "graphql_jwt.refresh_token.apps.RefreshTokenConfig" in django_settings.INSTALLED_APPS
    )


def revoke_user_refresh_token(user):
    if using_refresh_tokens():
        refresh_tokens = user.refresh_tokens.all()
        for refresh_token in refresh_tokens:
            try:
                refresh_token.revoke()
            except Exception:  # JSONWebTokenError
                pass


def flat_dict(dict_or_list):
    """
    if is dict, return list of dict keys,
    if is list, return the list
    """
    return list(dict_or_list.keys()) if isinstance(dict_or_list, dict) else dict_or_list


def normalize_fields(dict_or_list, extra_list):
    """
    helper merge settings defined fileds and
    other fields on mutations
    """
    if isinstance(dict_or_list, dict):
        for i in extra_list:
            dict_or_list[i] = "String"
        return dict_or_list
    else:
        return dict_or_list + extra_list


def get_classes(module: str | types.ModuleType, class_type: type | None = None) -> list[tuple[str, type]]:
    """
    Returns list of all classes od type `class_type` defined in the given module path.
    """
    if isinstance(module, str):
        module = import_module(module)
    if not class_type:
        return inspect.getmembers(module, inspect.isclass)
    return [item for item in inspect.getmembers(module, inspect.isclass) if issubclass(item[1], class_type)]


def camelize_form_errors(errors: dict) -> dict:
    """Camelize dict of django form errors"""
    if errors.get('__all__', False):
        errors['non_field_errors'] = errors.pop('__all__')
    return camelize(errors)  # type: ignore


def get_user_by_natural_key(username) -> AbstractBaseUser | None:
    """
    A difference approach from the original method (graphql_jwt.utils.get_user_by_natural_key)
    is using `select_related('status')`
    """
    UserModel = get_user_model()
    return UserModel._default_manager.select_related('status').filter(**{UserModel.USERNAME_FIELD: username}).first()  # type: ignore



from typing import Callable

from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q
from django.utils.module_loading import import_string

from .settings import graphql_auth_settings as app_settings
from .types import ExpectedErrorType

UserModel = get_user_model()


def get_user_by_email(email):
    """
    get user by email or by secondary email
    raise ObjectDoesNotExist
    """
    user = (
        UserModel._default_manager.select_related('status')
        .filter(Q(**{UserModel.EMAIL_FIELD: email}) | Q(status__secondary_email=email))  # type: ignore
        .first()
    )
    if user is None:
        raise ObjectDoesNotExist
    return user


def get_user_to_login(**kwargs):
    """
    get user by kwargs or secondary email
    to perform login
    raise ObjectDoesNotExist
    """
    if 'email' in kwargs.keys():
        lookup_filter = Q(email=kwargs['email'])
        if app_settings.ALLOW_LOGIN_WITH_SECONDARY_EMAIL:
            lookup_filter |= Q(status__secondary_email=kwargs['email'])
        user = UserModel._default_manager.select_related('status').filter(lookup_filter).first()
    else:
        user = UserModel._default_manager.select_related('status').filter(**kwargs).first()
    if user:
        return user
    else:
        raise ObjectDoesNotExist


def get_async_email_func() -> Callable | None:
    if app_settings.is_async_email:
        return import_string(app_settings.EMAIL_ASYNC_TASK)
    return None


async_email_func = get_async_email_func()


def get_output_error_type():
    if app_settings.CUSTOM_ERROR_TYPE and isinstance(app_settings.CUSTOM_ERROR_TYPE, str):
        return import_string(app_settings.CUSTOM_ERROR_TYPE)
    else:
        return ExpectedErrorType


OutputErrorType = get_output_error_type()



from smtplib import SMTPException

import graphene
from graphene.types.generic import GenericScalar
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import PasswordChangeForm, SetPasswordForm
from django.core.exceptions import ObjectDoesNotExist
from django.core.signing import BadSignature, SignatureExpired
from django.db import transaction
from django.utils.translation import gettext as _
from graphql_jwt.decorators import token_auth
from graphql_jwt.exceptions import JSONWebTokenError, JSONWebTokenExpired

from .bases import SuccessErrorsOutput
from .constants import Messages, TokenAction
from .decorators import password_confirmation_required, secondary_email_required, verification_required
from .exceptions import (
    EmailAlreadyInUseError,
    InvalidCredentialsError,
    PasswordAlreadySetError,
    TokenScopeError,
    UserAlreadyVerifiedError,
    UserNotVerifiedError,
    WrongUsageError,
)
from .forms import EmailForm, PasswordLessRegisterForm, RegisterForm, UpdateAccountForm
from .models import UserStatus
from .queries import UserNode
from .settings import graphql_auth_settings as app_settings
from .shortcuts import async_email_func, get_user_by_email, get_user_to_login
from .signals import user_registered, user_verified
from .utils import get_token_payload, revoke_user_refresh_token, using_refresh_tokens

UserModel = get_user_model()


class RegisterMixin(SuccessErrorsOutput):
    """
    Register user with fields defined in the settings.

    If the email field of the user model is part of the
    registration fields (default), check if there is
    no user with that email or as a secondary email.

    If it exists, it does not register the user,
    even if the email field is not defined as unique
    (default of the default django user model).

    When creating the user, it also creates a `UserStatus`
    related to that user, making it possible to track
    if the user is archived, verified and has a secondary
    email.

    Send account verification email.

    If allowed to not verified users login, return token.
    """

    if app_settings.ALLOW_LOGIN_NOT_VERIFIED:
        token = graphene.Field(graphene.String)
        if using_refresh_tokens():
            refresh_token = graphene.Field(graphene.String)

    form = PasswordLessRegisterForm if app_settings.ALLOW_PASSWORDLESS_REGISTRATION else RegisterForm

    @classmethod
    @token_auth
    def login_on_register(cls, root, info, **kwargs):
        return cls()

    @classmethod
    def resolve_mutation(cls, root, info, **kwargs):
        try:
            with transaction.atomic():
                f = cls.form(kwargs)
                if f.is_valid():
                    email = kwargs.get(UserModel.EMAIL_FIELD, False)  # type: ignore
                    UserStatus.clean_email(email)
                    user = f.save()
                    send_activation = app_settings.SEND_ACTIVATION_EMAIL is True and email
                    send_password_set = (
                        app_settings.ALLOW_PASSWORDLESS_REGISTRATION is True
                        and app_settings.SEND_PASSWORD_SET_EMAIL is True
                        and email
                    )
                    if send_activation:
                        if app_settings.is_async_email:
                            async_email_func(user.status.send_activation_email, (info,))  # type: ignore
                        else:
                            user.status.send_activation_email(info)

                    if send_password_set:
                        if async_email_func:
                            async_email_func(user.status.send_password_set_email, (info,))
                        else:
                            user.status.send_password_set_email(info)

                    user_registered.send(sender=cls, user=user)

                    if app_settings.ALLOW_LOGIN_NOT_VERIFIED:
                        payload = cls.login_on_register(root, info, password=kwargs.get("password1"), **kwargs)
                        return_value = {}
                        for field in cls._meta.fields:  # type: ignore
                            return_value[field] = getattr(payload, field)
                        return cls(**return_value)
                    return cls(success=True)
                else:
                    return cls(success=False, errors=f.errors)
        except EmailAlreadyInUseError:
            return cls(
                success=False,
                # if the email was set as a secondary email,
                # the RegisterForm will not catch it,
                # so we need to run UserStatus.clean_email(email)
                errors=Messages.EMAIL_IN_USE,
            )
        except SMTPException:
            return cls(success=False, errors=Messages.FAILED_SENDING_ACTIVATION_EMAIL)


class VerifyAccountMixin(SuccessErrorsOutput):
    """
    Verify user account.

    Receive the token that was sent by email.
    If the token is valid, make the user verified
    by making the `user.status.verified` field true.
    """

    @classmethod
    def resolve_mutation(cls, root, info, **kwargs):
        try:
            token = kwargs.get("token")
            UserStatus.verify(token)
            return cls(success=True)
        except UserAlreadyVerifiedError:
            return cls(success=False, errors=Messages.ALREADY_VERIFIED)
        except SignatureExpired:
            return cls(success=False, errors=Messages.EXPIRED_TOKEN)
        except (BadSignature, TokenScopeError):
            return cls(success=False, errors=Messages.INVALID_TOKEN)


class VerifySecondaryEmailMixin(SuccessErrorsOutput):
    """
    Verify user secondary email.

    Receive the token that was sent by email.
    User is already verified when using this mutation.

    If the token is valid, add the secondary email
    to `user.status.secondary_email` field.

    Note that until the secondary email is verified,
    it has not been saved anywhere beyond the token,
    so it can still be used to create a new account.
    After being verified, it will no longer be available.
    """

    @classmethod
    def resolve_mutation(cls, root, info, **kwargs):
        try:
            token = kwargs.get("token")
            UserStatus.verify_secondary_email(token)
            return cls(success=True)
        except EmailAlreadyInUseError:
            # while the token for secondary email was sent and the user haven't
            # verified, the email has not been saved to database and it was free.
            return cls(success=False, errors=Messages.EMAIL_IN_USE)
        except SignatureExpired:
            return cls(success=False, errors=Messages.EXPIRED_TOKEN)
        except (BadSignature, TokenScopeError):
            return cls(success=False, errors=Messages.INVALID_TOKEN)


class ResendActivationEmailMixin(SuccessErrorsOutput):
    """
    Sends activation email.

    It is called resend because theoretically
    the first activation email was sent when
    the user registered.

    If there is no user with the requested email,
    a successful response is returned.
    """

    form = EmailForm

    @classmethod
    def resolve_mutation(cls, root, info, **kwargs):
        try:
            email = kwargs.get("email")
            f = cls.form({"email": email})
            if f.is_valid():
                user = get_user_by_email(email)
                if async_email_func:
                    async_email_func(user.status.resend_activation_email, (info,))  # type: ignore
                else:
                    user.status.resend_activation_email(info)  # type: ignore
                return cls(success=True)
            return cls(success=False, errors=f.errors)
        except ObjectDoesNotExist:
            return cls(success=True)  # even if user is not registered
        except SMTPException:
            return cls(success=False, errors=Messages.EMAIL_FAIL)
        except UserAlreadyVerifiedError:
            return cls(success=False, errors=Messages.ALREADY_VERIFIED)


class SendPasswordResetEmailMixin(SuccessErrorsOutput):
    """
    Send password reset email.

    For non verified users, send an activation
    email instead.

    Accepts both primary and secondary email.

    If there is no user with the requested email,
    a successful response is returned.
    """

    form = EmailForm

    @classmethod
    def resolve_mutation(cls, root, info, **kwargs):
        email = kwargs.get("email")
        try:
            f = cls.form({"email": email})
            if f.is_valid():
                user = get_user_by_email(email)
                if async_email_func:
                    async_email_func(user.status.send_password_reset_email, (info, [email]))  # type: ignore
                else:
                    user.status.send_password_reset_email(info, [email])  # type: ignore
                return cls(success=True)
            return cls(success=False, errors=f.errors)
        except ObjectDoesNotExist:
            return cls(success=True)  # even if user is not registered
        except SMTPException:
            return cls(success=False, errors=Messages.EMAIL_FAIL)

    @classmethod
    def _resend_activation_email(cls, info, user):  # pragma: no cover
        try:
            if async_email_func:
                async_email_func(user.status.resend_activation_email, (info,))  # type: ignore
            else:
                user.status.resend_activation_email(info)  # type: ignore
            return cls(success=False, errors=Messages.NOT_VERIFIED_PASSWORD_RESET)
        except SMTPException:
            return cls(success=False, errors=Messages.EMAIL_FAIL)


class PasswordResetMixin(SuccessErrorsOutput):
    """
    Change user password without old password.

    Receive the token that was sent by email.

    If token and new passwords are valid, update
    user password and in case of using refresh
    tokens, revoke all of them.

    Also, if user has not been verified yet, verify it.
    """

    form = SetPasswordForm

    @classmethod
    def resolve_mutation(cls, root, info, **kwargs):
        try:
            token = kwargs.pop("token")
            payload = get_token_payload(
                token,
                TokenAction.PASSWORD_RESET,
                app_settings.EXPIRATION_PASSWORD_RESET_TOKEN,
            )
            user = UserModel._default_manager.get(**payload)
            f = cls.form(user, kwargs)
            if f.is_valid():
                revoke_user_refresh_token(user)
                user = f.save()

                if user.status.verified is False:  # type: ignore
                    user.status.verified = True  # type: ignore
                    user.status.save(update_fields=["verified"])  # type: ignore
                    user_verified.send(sender=cls, user=user)

                return cls(success=True)
            return cls(success=False, errors=f.errors)
        except SignatureExpired:
            return cls(success=False, errors=Messages.EXPIRED_TOKEN)
        except (BadSignature, TokenScopeError):
            return cls(success=False, errors=Messages.INVALID_TOKEN)


class PasswordSetMixin(SuccessErrorsOutput):
    """
    Set user password - for passwordless registration

    Receive the token that was sent by email.

    If token and new passwords are valid, set
    user password and in case of using refresh
    tokens, revoke all of them.

    Also, if user has not been verified yet, verify it.
    """

    form = SetPasswordForm

    @classmethod
    def resolve_mutation(cls, root, info, **kwargs):
        try:
            token = kwargs.pop("token")
            payload = get_token_payload(
                token,
                TokenAction.PASSWORD_SET,
                app_settings.EXPIRATION_PASSWORD_SET_TOKEN,
            )
            user = UserModel._default_manager.get(**payload)
            f = cls.form(user, kwargs)
            if f.is_valid():
                # Check if user has already set a password
                if user.has_usable_password():
                    raise PasswordAlreadySetError
                revoke_user_refresh_token(user)
                user = f.save()

                if user.status.verified is False:  # type: ignore
                    user.status.verified = True  # type: ignore
                    user.status.save(update_fields=["verified"])  # type: ignore

                return cls(success=True)
            return cls(success=False, errors=f.errors)
        except SignatureExpired:
            return cls(success=False, errors=Messages.EXPIRED_TOKEN)
        except (BadSignature, TokenScopeError):
            return cls(success=False, errors=Messages.INVALID_TOKEN)
        except PasswordAlreadySetError:
            return cls(success=False, errors=Messages.PASSWORD_ALREADY_SET)


class ObtainJSONWebTokenMixin(SuccessErrorsOutput):
    """
    Obtain JSON web token for given user.

    Allow to perform login with different fields,
    and secondary email if set. The fields are
    defined on settings.

    Not verified users can login by default. This
    can be changes on settings.

    If user is archived, make it unarchive and
    return `unarchiving=True` on output.
    """

    refresh_expires_in = graphene.Int()
    payload = GenericScalar()
    user = graphene.Field(UserNode)
    unarchiving = graphene.Boolean(default_value=False)

    user_to_login = None

    @classmethod
    def resolve(cls, root, info, **kwargs):
        user = cls.user_to_login if cls.user_to_login == info.context.user else info.context.user
        if user.status.archived is True:  # unarchive on login # type: ignore
            UserStatus.unarchive(user)
            unarchiving = True
        else:
            unarchiving = False
        return cls(user=info.context.user, unarchiving=unarchiving)

    @classmethod
    def resolve_mutation(cls, root, info, **kwargs):
        if len(kwargs.items()) != 2:
            raise WrongUsageError(
                "Must login with password and one of the following fields %s." % (app_settings.LOGIN_ALLOWED_FIELDS)
            )

        try:
            final_kwargs = None
            USERNAME_FIELD = UserModel.USERNAME_FIELD  # type: ignore

            # extract USERNAME_FIELD to use in query
            if USERNAME_FIELD in kwargs:
                data_to_login = {USERNAME_FIELD: kwargs[USERNAME_FIELD]}
                final_kwargs = kwargs
                password = kwargs.get("password")
            else:  # use what is left to query
                password = kwargs.pop("password")
                query_field, query_value = kwargs.popitem()
                data_to_login = {query_field: query_value}

            cls.user_to_login = get_user_to_login(**data_to_login)

            if not final_kwargs:
                final_kwargs = {
                    "password": password,
                    USERNAME_FIELD: getattr(cls.user_to_login, USERNAME_FIELD),
                }

            if cls.user_to_login.status.verified or app_settings.ALLOW_LOGIN_NOT_VERIFIED:  # type: ignore
                return cls.parent_resolve(root, info, **final_kwargs)  # type: ignore
            else:
                raise UserNotVerifiedError
        except (JSONWebTokenError, ObjectDoesNotExist, InvalidCredentialsError):
            return cls(success=False, errors=Messages.INVALID_CREDENTIALS)
        except UserNotVerifiedError:
            return cls(success=False, errors=Messages.NOT_VERIFIED)


class ArchiveOrDeleteMixin(SuccessErrorsOutput):
    @classmethod
    @verification_required
    @password_confirmation_required
    def resolve_mutation(cls, root, info, *args, **kwargs):
        user = info.context.user
        cls.resolve_action(user, root=root, info=info)
        return cls(success=True)

    @classmethod
    def resolve_action(cls, user, *args, **kwargs):
        raise NotImplementedError


class ArchiveAccountMixin(ArchiveOrDeleteMixin):
    """
    Archive account and revoke refresh tokens.

    User must be verified and confirm password.
    """

    @classmethod
    def resolve_action(cls, user, *args, **kwargs):
        UserStatus.archive(user)
        revoke_user_refresh_token(user=user)


class DeleteAccountMixin(ArchiveOrDeleteMixin):
    """
    Delete account permanently or make `user.is_active=False`.

    The behavior is defined on settings.
    Anyway user refresh tokens are revoked.

    User must be verified and confirm password.
    """

    @classmethod
    def resolve_action(cls, user, *args, **kwargs):
        if app_settings.ALLOW_DELETE_ACCOUNT:
            revoke_user_refresh_token(user=user)
            user.delete()
        else:
            user.is_active = False
            user.save(update_fields=["is_active"])
            revoke_user_refresh_token(user=user)


class PasswordChangeMixin(SuccessErrorsOutput):
    """
    Change account password when user knows the old password.

    A new token and refresh token are sent. User must be verified.
    """

    token = graphene.Field(graphene.String)
    if using_refresh_tokens():
        refresh_token = graphene.Field(graphene.String)

    form = PasswordChangeForm

    @classmethod
    @token_auth
    def login_on_password_change(cls, root, info, **kwargs):
        return cls()

    @classmethod
    @verification_required
    @password_confirmation_required
    def resolve_mutation(cls, root, info, **kwargs):
        user = info.context.user
        f = cls.form(user, kwargs)
        if f.is_valid():
            revoke_user_refresh_token(user)
            user = f.save()
            payload = cls.login_on_password_change(
                root,
                info,
                password=kwargs.get("new_password1"),
                **{user.USERNAME_FIELD: getattr(user, user.USERNAME_FIELD)}  # type: ignore
            )
            return_value = {}
            for field in cls._meta.fields:  # type: ignore
                return_value[field] = getattr(payload, field)
            return cls(**return_value)
        else:
            return cls(success=False, errors=f.errors)


class UpdateAccountMixin(SuccessErrorsOutput):
    """
    Update user model fields, defined on settings.

    User must be verified.
    """

    form = UpdateAccountForm

    @classmethod
    @verification_required
    def resolve_mutation(cls, root, info, **kwargs):
        user = info.context.user
        fields = cls.form.Meta.fields
        for field in fields:
            if field not in kwargs:
                kwargs[field] = getattr(user, field)
        f = cls.form(kwargs, instance=user)
        if f.is_valid():
            f.save()
            return cls(success=True)
        return cls(success=False, errors=f.errors)


class VerifyOrRefreshOrRevokeTokenMixin(SuccessErrorsOutput):
    """
    Same as `grapgql_jwt` implementation,
    with change exception error message from "Error decoding signature" to "Invalid token.".
    """

    @classmethod
    def resolve_mutation(cls, root, info, **kwargs):
        try:
            return cls.parent_resolve(root, info, **kwargs)  # type: ignore
        except JSONWebTokenExpired:
            return cls(success=False, errors=Messages.EXPIRED_TOKEN)
        except JSONWebTokenError as e:
            if str(e).endswith('expired'):
                return cls(success=False, errors=Messages.EXPIRED_TOKEN)
            return cls(success=False, errors=Messages.INVALID_TOKEN)


class SendSecondaryEmailActivationMixin(SuccessErrorsOutput):
    """
    Send activation to secondary email.

    User must be verified and confirm password.
    """

    @classmethod
    @verification_required
    @password_confirmation_required
    def resolve_mutation(cls, root, info, **kwargs):
        try:
            email = kwargs.get("email")
            f = EmailForm({"email": email})
            if f.is_valid():
                user = info.context.user
                if async_email_func:
                    async_email_func(user.status.send_secondary_email_activation, (info, email))
                else:
                    user.status.send_secondary_email_activation(info, email)
                return cls(success=True)
            return cls(success=False, errors=f.errors)
        except EmailAlreadyInUseError:
            return cls(success=False, errors=Messages.EMAIL_IN_USE)
        except SMTPException:
            return cls(success=False, errors=Messages.EMAIL_FAIL)


class SwapEmailsMixin(SuccessErrorsOutput):
    """
    Swap between primary and secondary emails.

    Require password confirmation.
    """

    @classmethod
    @secondary_email_required
    @password_confirmation_required
    def resolve_mutation(cls, root, info, **kwargs):
        info.context.user.status.swap_emails()
        return cls(success=True)


class RemoveSecondaryEmailMixin(SuccessErrorsOutput):
    """
    Remove user secondary email.

    Require password confirmation.
    """

    @classmethod
    @secondary_email_required
    @password_confirmation_required
    def resolve_mutation(cls, root, info, **kwargs):
        info.context.user.status.remove_secondary_email()
        return cls(success=True)



"""
Settings for graphql_auth are all namespaced in the GRAPHQL_AUTH setting.
For example your project's `settings.py` file might look like this:
GRAPHQL_AUTH = {
    'LOGIN_ALLOWED_FIELDS': ['email', 'username'],
    'SEND_ACTIVATION_EMAIL': True,
}
This module provides the `graphql_auth_settings` object, that is used to access
Graphene settings, checking for user settings first, then falling
back to the defaults.
"""

from datetime import timedelta

from django.conf import settings as django_settings
from django.test.signals import setting_changed

# Copied shamelessly from Graphene / Django REST Framework

DEFAULTS = {
    # if allow to login without verification,
    # the register mutation will return a token
    'ALLOW_LOGIN_NOT_VERIFIED': True,
    # mutations fields options
    'LOGIN_ALLOWED_FIELDS': ['email', 'username'],
    'ALLOW_LOGIN_WITH_SECONDARY_EMAIL': True,
    # required fields on register, plus password1 and password2,
    # can be a dict like UPDATE_MUTATION_FIELDS setting
    'REGISTER_MUTATION_FIELDS': ['email', 'username'],
    'REGISTER_MUTATION_FIELDS_OPTIONAL': [],
    # optional fields on update account, can be list of fields
    'UPDATE_MUTATION_FIELDS': {'first_name': 'String', 'last_name': 'String'},
    # tokens
    'EXPIRATION_ACTIVATION_TOKEN': timedelta(days=7),
    'EXPIRATION_PASSWORD_RESET_TOKEN': timedelta(hours=1),
    'EXPIRATION_SECONDARY_EMAIL_ACTIVATION_TOKEN': timedelta(hours=1),
    'EXPIRATION_PASSWORD_SET_TOKEN': timedelta(hours=1),
    # email stuff
    'EMAIL_FROM': getattr(django_settings, 'DEFAULT_FROM_EMAIL', 'test@email.com'),
    'SEND_ACTIVATION_EMAIL': True,
    # client: example.com/activate/token
    'ACTIVATION_PATH_ON_EMAIL': 'activate',
    'ACTIVATION_SECONDARY_EMAIL_PATH_ON_EMAIL': 'activate',
    # client: example.com/password-set/token
    'PASSWORD_SET_PATH_ON_EMAIL': 'password-set',
    # client: example.com/password-reset/token
    'PASSWORD_RESET_PATH_ON_EMAIL': 'password-reset',
    # email subjects templates
    'EMAIL_SUBJECT_ACTIVATION': 'email/activation_subject.txt',
    'EMAIL_SUBJECT_ACTIVATION_RESEND': 'email/activation_subject.txt',
    'EMAIL_SUBJECT_SECONDARY_EMAIL_ACTIVATION': 'email/activation_subject.txt',
    'EMAIL_SUBJECT_PASSWORD_SET': 'email/password_set_subject.txt',
    'EMAIL_SUBJECT_PASSWORD_RESET': 'email/password_reset_subject.txt',
    # email templates
    'EMAIL_TEMPLATE_ACTIVATION': 'email/activation_email.html',
    'EMAIL_TEMPLATE_ACTIVATION_RESEND': 'email/activation_email.html',
    'EMAIL_TEMPLATE_SECONDARY_EMAIL_ACTIVATION': 'email/activation_email.html',
    'EMAIL_TEMPLATE_PASSWORD_SET': 'email/password_set_email.html',
    'EMAIL_TEMPLATE_PASSWORD_RESET': 'email/password_reset_email.html',
    'EMAIL_TEMPLATE_VARIABLES': {},
    # query stuff
    'USER_NODE_EXCLUDE_FIELDS': ['password', 'is_superuser'],
    'USER_NODE_FILTER_FIELDS': {
        'email': ['exact'],
        'username': ['exact', 'icontains', 'istartswith'],
        'is_active': ['exact'],
        'status__archived': ['exact'],
        'status__verified': ['exact'],
        'status__secondary_email': ['exact'],
    },
    # turn is_active to False instead
    'ALLOW_DELETE_ACCOUNT': False,
    # string path for email function wrapper, see the testproject example
    'EMAIL_ASYNC_TASK': False,
    # # mutation error type
    'CUSTOM_ERROR_TYPE': None,
    # registration with no password
    'ALLOW_PASSWORDLESS_REGISTRATION': False,
    'SEND_PASSWORD_SET_EMAIL': False,
}


class GraphQLAuthSettings(object):
    """
    A settings object, that allows API settings to be accessed as properties.
    For example:
        from graphql_auth.settings import settings
        print(settings)
    """

    def __init__(self, user_settings=None, defaults=None):
        if user_settings:
            self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(django_settings, 'GRAPHQL_AUTH', {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError(f"Invalid graphql_auth setting: {attr}")

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Cache the result
        setattr(self, attr, val)
        return val

    @property
    def is_async_email(self) -> bool:
        try:
            return self._is_declared_async_email
        except AttributeError:
            self._is_declared_async_email = bool(self.EMAIL_ASYNC_TASK and isinstance(self.EMAIL_ASYNC_TASK, str))
            return self._is_declared_async_email


graphql_auth_settings = GraphQLAuthSettings(None, DEFAULTS)


def reload_graphql_auth_settings(*args, **kwargs):
    global graphql_auth_settings
    setting, value = kwargs['setting'], kwargs['value']
    if setting == 'GRAPHQL_AUTH':
        for key in DEFAULTS.keys():
            try:
                delattr(graphql_auth_settings, key)
            except AttributeError:
                pass
        try:
            delattr(graphql_auth_settings, '_is_declared_async_email')
        except AttributeError:
            pass
        graphql_auth_settings._user_settings = value


setting_changed.connect(reload_graphql_auth_settings)



from typing import Any, Collection

from django.utils.translation import gettext as _
from graphql import GraphQLError, GraphQLErrorExtensions
from graphql.language.ast import Node
from graphql.language.source import Source

from graphql_auth.constants import Messages


class GraphQLAuthError(GraphQLError):
    default_message: str = _('Not described message.')
    _extensions: dict[str, Any] = {}

    def __init__(
        self,
        message: str | None = None,
        nodes: Collection[Node] | Node | None = None,
        source: Source | None = None,
        positions: Collection[int] | None = None,
        path: Collection[str | int] | None = None,
        original_error: Exception | None = None,
        extensions: GraphQLErrorExtensions | None = None,
    ) -> None:
        if message is None:
            message = self.default_message
        if not extensions and self._extensions:
            extensions = self._extensions
        super().__init__(message, nodes, source, positions, path, original_error, extensions)


class UserAlreadyVerifiedError(GraphQLAuthError):
    default_message = Messages.ALREADY_VERIFIED['message']
    _extensions = Messages.ALREADY_VERIFIED


class InvalidCredentialsError(GraphQLAuthError):
    default_message = _("Invalid credentials.")
    _extensions = Messages.INVALID_CREDENTIALS


class UserNotVerifiedError(GraphQLAuthError):
    default_message = _("User is not verified.")
    _extensions = Messages.NOT_VERIFIED


class EmailAlreadyInUseError(GraphQLAuthError):
    default_message = _("This email is already in use.")
    _extensions = Messages.EMAIL_IN_USE


class TokenScopeError(GraphQLAuthError):
    default_message = _("This token if for something else.")


class PasswordAlreadySetError(GraphQLAuthError):
    default_message = Messages.PASSWORD_ALREADY_SET['message']
    _extensions = Messages.PASSWORD_ALREADY_SET


class WrongUsageError(GraphQLAuthError):
    """internal exception"""

    default_message = _("Wrong usage, check your code!.")


class InvalidEmailAddressError(GraphQLAuthError):
    default_message = _('Invalid email address')
    _extensions = {'message': default_message, 'code': 'invalid-email-address'}



from django.utils.translation import gettext as _


class Messages:
    INVALID_PASSWORD = {"message": _("Invalid password."), "code": "invalid_password"}
    UNAUTHENTICATED = {"message": _("Unauthenticated."), "code": "unauthenticated"}
    INVALID_TOKEN = {"message": _("Invalid token."), "code": "invalid_token"}
    EXPIRED_TOKEN = {"message": _("Expired token."), "code": "expired_token"}
    ALREADY_VERIFIED = {"message": _("Account already verified."), "code": "already_verified"}
    EMAIL_FAIL = {"message": _("Failed to send email."), "code": "email_fail"}
    INVALID_CREDENTIALS = {
        "message": _("Please, enter valid credentials."),
        "code": "invalid_credentials",
    }
    NOT_VERIFIED = {"message": _("Please verify your account."), "code": "not_verified"}
    NOT_VERIFIED_PASSWORD_RESET = {
        "message": _("Verify your account first. A new verification email was sent."),
        "code": "not_verified",
    }
    EMAIL_IN_USE = {"email": _("A user with that email already exists."), "code": "email_in_use"}
    SECONDARY_EMAIL_REQUIRED = {
        "message": _("You need to setup a secondary email to proceed."),
        "code": "secondary_email_required",
    }
    PASSWORD_ALREADY_SET = {
        "message": _("Password already set for account."),
        "code": "password_already_set",
    }
    INVALID_REGISTRATION_DATA_MESSAGE = _("Invalid registration data.")
    FAILED_SENDING_ACTIVATION_EMAIL = {
        'message': _("User account created but could not send activation email."),
        'code': 'send_activation_email_failed',
    }
    FAILED_PASSWORD_CHANGE_MESSAGE = _("Password change failed.")
    INVALID_EMAIL_ADDRESS_MESSAGE = _("Invalid email address.")


class TokenAction(object):
    ACTIVATION = "activation"
    PASSWORD_RESET = "password_reset"
    ACTIVATION_SECONDARY_EMAIL = "activation_secondary_email"
    PASSWORD_SET = "password_set"



import graphene
import graphql_jwt
from graphene.types.generic import GenericScalar
from graphql_jwt.mixins import JSONWebTokenMixin
from graphql_jwt.settings import jwt_settings

from .bases import DynamicArgsMixin, MutationMixin
from .mixins import (
    ArchiveAccountMixin,
    DeleteAccountMixin,
    ObtainJSONWebTokenMixin,
    PasswordChangeMixin,
    PasswordResetMixin,
    PasswordSetMixin,
    RegisterMixin,
    RemoveSecondaryEmailMixin,
    ResendActivationEmailMixin,
    SendPasswordResetEmailMixin,
    SendSecondaryEmailActivationMixin,
    SwapEmailsMixin,
    UpdateAccountMixin,
    VerifyAccountMixin,
    VerifyOrRefreshOrRevokeTokenMixin,
    VerifySecondaryEmailMixin,
)
from .settings import graphql_auth_settings as app_settings
from .utils import normalize_fields


class Register(MutationMixin, DynamicArgsMixin, RegisterMixin, graphene.Mutation):
    _required_args = normalize_fields(
        app_settings.REGISTER_MUTATION_FIELDS,
        [] if app_settings.ALLOW_PASSWORDLESS_REGISTRATION else ["password1", "password2"],
    )
    _args = app_settings.REGISTER_MUTATION_FIELDS_OPTIONAL
    __doc__ = RegisterMixin.__doc__


class VerifyAccount(MutationMixin, DynamicArgsMixin, VerifyAccountMixin, graphene.Mutation):
    _required_args = ["token"]
    __doc__ = VerifyAccountMixin.__doc__


class ResendActivationEmail(MutationMixin, DynamicArgsMixin, ResendActivationEmailMixin, graphene.Mutation):
    _required_args = ["email"]
    __doc__ = ResendActivationEmailMixin.__doc__


class SendPasswordResetEmail(MutationMixin, DynamicArgsMixin, SendPasswordResetEmailMixin, graphene.Mutation):
    _required_args = ["email"]
    __doc__ = SendPasswordResetEmailMixin.__doc__


class SendSecondaryEmailActivation(
    MutationMixin, DynamicArgsMixin, SendSecondaryEmailActivationMixin, graphene.Mutation
):
    _required_args = ["email", "password"]
    __doc__ = SendSecondaryEmailActivationMixin.__doc__


class VerifySecondaryEmail(MutationMixin, DynamicArgsMixin, VerifySecondaryEmailMixin, graphene.Mutation):
    _required_args = ["token"]
    __doc__ = VerifySecondaryEmailMixin.__doc__


class SwapEmails(MutationMixin, DynamicArgsMixin, SwapEmailsMixin, graphene.Mutation):
    _required_args = ["password"]
    __doc__ = SwapEmailsMixin.__doc__


class RemoveSecondaryEmail(MutationMixin, DynamicArgsMixin, RemoveSecondaryEmailMixin, graphene.Mutation):
    _required_args = ["password"]
    __doc__ = RemoveSecondaryEmailMixin.__doc__


class PasswordSet(MutationMixin, PasswordSetMixin, DynamicArgsMixin, graphene.Mutation):
    _required_args = ["token", "new_password1", "new_password2"]
    __doc__ = PasswordSetMixin.__doc__


class PasswordReset(MutationMixin, DynamicArgsMixin, PasswordResetMixin, graphene.Mutation):
    _required_args = ["token", "new_password1", "new_password2"]
    __doc__ = PasswordResetMixin.__doc__


class ObtainJSONWebToken(MutationMixin, ObtainJSONWebTokenMixin, graphql_jwt.JSONWebTokenMutation):
    __doc__ = ObtainJSONWebTokenMixin.__doc__

    @classmethod
    def Field(cls, *args, **kwargs):
        cls._meta.arguments.update({"password": graphene.String(required=True)})
        for field in app_settings.LOGIN_ALLOWED_FIELDS:
            cls._meta.arguments.update({field: graphene.String()})
        if not jwt_settings.JWT_HIDE_TOKEN_FIELDS:
            cls._meta.fields['token'] = graphene.Field(graphene.String, required=False)
            if jwt_settings.JWT_LONG_RUNNING_REFRESH_TOKEN:
                cls._meta.fields['refresh_token'] = graphene.Field(graphene.String, required=False)
        return super(JSONWebTokenMixin, cls).Field(*args, **kwargs)


class ArchiveAccount(MutationMixin, ArchiveAccountMixin, DynamicArgsMixin, graphene.Mutation):
    _required_args = ["password"]
    __doc__ = ArchiveAccountMixin.__doc__


class DeleteAccount(MutationMixin, DeleteAccountMixin, DynamicArgsMixin, graphene.Mutation):
    _required_args = ["password"]
    __doc__ = DeleteAccountMixin.__doc__


class PasswordChange(MutationMixin, PasswordChangeMixin, DynamicArgsMixin, graphene.Mutation):
    _required_args = ["old_password", "new_password1", "new_password2"]
    __doc__ = PasswordChangeMixin.__doc__


class UpdateAccount(MutationMixin, DynamicArgsMixin, UpdateAccountMixin, graphene.Mutation):
    _args = app_settings.UPDATE_MUTATION_FIELDS
    __doc__ = UpdateAccountMixin.__doc__


class VerifyToken(MutationMixin, VerifyOrRefreshOrRevokeTokenMixin, graphql_jwt.Verify):
    payload = GenericScalar(required=False)
    __doc__ = VerifyOrRefreshOrRevokeTokenMixin.__doc__


class RefreshToken(MutationMixin, VerifyOrRefreshOrRevokeTokenMixin, graphql_jwt.Refresh):
    payload = GenericScalar(required=False)
    __doc__ = VerifyOrRefreshOrRevokeTokenMixin.__doc__

    @classmethod
    def Field(cls, *args, **kwargs):
        if not jwt_settings.JWT_HIDE_TOKEN_FIELDS:
            cls._meta.fields["token"] = graphene.Field(graphene.String, required=False)

            if jwt_settings.JWT_LONG_RUNNING_REFRESH_TOKEN:
                cls._meta.fields["refresh_token"] = graphene.Field(graphene.String, required=False)

        return super(JSONWebTokenMixin, cls).Field(*args, **kwargs)  # type: ignore


class RevokeToken(MutationMixin, VerifyOrRefreshOrRevokeTokenMixin, graphql_jwt.Revoke):
    revoked = graphene.Int(required=False)
    __doc__ = VerifyOrRefreshOrRevokeTokenMixin.__doc__



import graphene


class CountableConnection(graphene.relay.Connection):
    class Meta:
        abstract = True

    total_count = graphene.Int()

    def resolve_total_count(self, info, **kwargs) -> int:
        return self.iterable.count()  # type: ignore



from typing import Any

import graphene
from django.contrib.auth import get_user_model
from graphene_django.filter.fields import DjangoFilterConnectionField
from graphene_django.types import DjangoObjectType

from .connection import CountableConnection
from .settings import graphql_auth_settings as app_settings

UserModel = get_user_model()


class UserNode(DjangoObjectType):
    class Meta:
        model = UserModel
        filter_fields = app_settings.USER_NODE_FILTER_FIELDS
        exclude = app_settings.USER_NODE_EXCLUDE_FIELDS
        interfaces = (graphene.relay.Node,)
        connection_class = CountableConnection

    pk = graphene.Int()
    archived = graphene.Boolean()
    verified = graphene.Boolean()
    secondary_email = graphene.String()

    def resolve_pk(self, info):
        return self.pk

    def resolve_archived(self, info):
        return self.status.archived  # type: ignore

    def resolve_verified(self, info):
        return self.status.verified  # type: ignore

    def resolve_secondary_email(self, info):
        return self.status.secondary_email  # type: ignore

    @classmethod
    def get_queryset(cls, queryset, info):
        return queryset.select_related('status')

    @classmethod
    def get_node(cls, info, id) -> Any:
        """
        Allows only staff users to access the node.
        """
        user = info.context.user
        if user.is_authenticated and user.is_staff:
            return super().get_node(info, id)
        return None


class UserQuery(graphene.ObjectType):
    user = graphene.relay.Node.Field(UserNode)
    users = DjangoFilterConnectionField(UserNode)

    def resolve_users(self, info, **kwargs):
        """
        Allows only staff users to get users list.
        """
        user = info.context.user
        if user.is_authenticated and user.is_staff:
            return UserModel.objects.all()
        return UserModel.objects.none()


class MeQuery(graphene.ObjectType):
    me = graphene.Field(UserNode)

    def resolve_me(self, info):
        user = info.context.user
        if user.is_authenticated:
            return user
        return None



import graphene
import graphql_jwt
from graphene.types.generic import GenericScalar
from graphql_jwt.mixins import JSONWebTokenMixin, RefreshMixin
from graphql_jwt.settings import jwt_settings

from .bases import DynamicInputMixin, RelayMutationMixin
from .mixins import (
    ArchiveAccountMixin,
    DeleteAccountMixin,
    ObtainJSONWebTokenMixin,
    PasswordChangeMixin,
    PasswordResetMixin,
    PasswordSetMixin,
    RegisterMixin,
    RemoveSecondaryEmailMixin,
    ResendActivationEmailMixin,
    SendPasswordResetEmailMixin,
    SendSecondaryEmailActivationMixin,
    SwapEmailsMixin,
    UpdateAccountMixin,
    VerifyAccountMixin,
    VerifyOrRefreshOrRevokeTokenMixin,
    VerifySecondaryEmailMixin,
)
from .settings import graphql_auth_settings as app_settings
from .utils import normalize_fields


class Register(RelayMutationMixin, DynamicInputMixin, RegisterMixin, graphene.ClientIDMutation):
    _required_inputs = normalize_fields(
        app_settings.REGISTER_MUTATION_FIELDS,
        [] if app_settings.ALLOW_PASSWORDLESS_REGISTRATION else ["password1", "password2"],
    )
    _inputs = app_settings.REGISTER_MUTATION_FIELDS_OPTIONAL
    __doc__ = RegisterMixin.__doc__


class VerifyAccount(RelayMutationMixin, DynamicInputMixin, VerifyAccountMixin, graphene.ClientIDMutation):
    _required_inputs = ["token"]
    __doc__ = VerifyAccountMixin.__doc__


class ResendActivationEmail(
    RelayMutationMixin, DynamicInputMixin, ResendActivationEmailMixin, graphene.ClientIDMutation
):
    _required_inputs = ["email"]
    __doc__ = ResendActivationEmailMixin.__doc__


class SendPasswordResetEmail(
    RelayMutationMixin, DynamicInputMixin, SendPasswordResetEmailMixin, graphene.ClientIDMutation
):
    _required_inputs = ["email"]
    __doc__ = SendPasswordResetEmailMixin.__doc__


class SendSecondaryEmailActivation(
    RelayMutationMixin,
    DynamicInputMixin,
    SendSecondaryEmailActivationMixin,
    graphene.ClientIDMutation,
):
    _required_inputs = ["email", "password"]
    __doc__ = SendSecondaryEmailActivationMixin.__doc__


class VerifySecondaryEmail(RelayMutationMixin, DynamicInputMixin, VerifySecondaryEmailMixin, graphene.ClientIDMutation):
    _required_inputs = ["token"]
    __doc__ = VerifySecondaryEmailMixin.__doc__


class SwapEmails(RelayMutationMixin, DynamicInputMixin, SwapEmailsMixin, graphene.ClientIDMutation):
    _required_inputs = ["password"]
    __doc__ = SwapEmailsMixin.__doc__


class RemoveSecondaryEmail(RelayMutationMixin, DynamicInputMixin, RemoveSecondaryEmailMixin, graphene.ClientIDMutation):
    _required_inputs = ["password"]
    __doc__ = RemoveSecondaryEmailMixin.__doc__


class PasswordSet(RelayMutationMixin, DynamicInputMixin, PasswordSetMixin, graphene.ClientIDMutation):
    _required_inputs = ["token", "new_password1", "new_password2"]
    __doc__ = PasswordSetMixin.__doc__


class PasswordReset(RelayMutationMixin, DynamicInputMixin, PasswordResetMixin, graphene.ClientIDMutation):
    _required_inputs = ["token", "new_password1", "new_password2"]
    __doc__ = PasswordResetMixin.__doc__


class ObtainJSONWebToken(RelayMutationMixin, ObtainJSONWebTokenMixin, graphql_jwt.relay.JSONWebTokenMutation):
    __doc__ = ObtainJSONWebTokenMixin.__doc__

    @classmethod
    def Field(cls, *args, **kwargs):
        cls._meta.arguments["input"]._meta.fields.update(
            {"password": graphene.InputField(graphene.String, required=True)}
        )
        for field in app_settings.LOGIN_ALLOWED_FIELDS:
            cls._meta.arguments["input"]._meta.fields.update({field: graphene.InputField(graphene.String)})
        if not jwt_settings.JWT_HIDE_TOKEN_FIELDS:
            cls._meta.fields['token'] = graphene.Field(graphene.String, required=False)
            if jwt_settings.JWT_LONG_RUNNING_REFRESH_TOKEN:
                cls._meta.fields['refresh_token'] = graphene.Field(graphene.String, required=False)
        return super(JSONWebTokenMixin, cls).Field(*args, **kwargs)


class ArchiveAccount(RelayMutationMixin, ArchiveAccountMixin, DynamicInputMixin, graphene.ClientIDMutation):
    _required_inputs = ["password"]
    __doc__ = ArchiveAccountMixin.__doc__


class DeleteAccount(RelayMutationMixin, DeleteAccountMixin, DynamicInputMixin, graphene.ClientIDMutation):
    _required_inputs = ["password"]
    __doc__ = DeleteAccountMixin.__doc__


class PasswordChange(RelayMutationMixin, PasswordChangeMixin, DynamicInputMixin, graphene.ClientIDMutation):
    _required_inputs = ["old_password", "new_password1", "new_password2"]
    __doc__ = PasswordChangeMixin.__doc__


class UpdateAccount(RelayMutationMixin, DynamicInputMixin, UpdateAccountMixin, graphene.ClientIDMutation):
    _inputs = app_settings.UPDATE_MUTATION_FIELDS
    __doc__ = UpdateAccountMixin.__doc__


class VerifyToken(RelayMutationMixin, VerifyOrRefreshOrRevokeTokenMixin, graphql_jwt.relay.Verify):
    payload = GenericScalar(required=False)
    __doc__ = VerifyOrRefreshOrRevokeTokenMixin.__doc__

    class Input:
        token = graphene.String(required=True)


class RefreshToken(RelayMutationMixin, VerifyOrRefreshOrRevokeTokenMixin, graphql_jwt.relay.Refresh):
    payload = GenericScalar(required=False)
    __doc__ = VerifyOrRefreshOrRevokeTokenMixin.__doc__

    class Input(RefreshMixin.Fields):
        """Refresh Input"""

    @classmethod
    def Field(cls, *args, **kwargs):
        if not jwt_settings.JWT_HIDE_TOKEN_FIELDS:
            cls._meta.fields["token"] = graphene.Field(graphene.String, required=False)

            if jwt_settings.JWT_LONG_RUNNING_REFRESH_TOKEN:
                cls._meta.fields["refresh_token"] = graphene.Field(graphene.String, required=False)

        return super(JSONWebTokenMixin, cls).Field(*args, **kwargs)  # type: ignore


class RevokeToken(RelayMutationMixin, VerifyOrRefreshOrRevokeTokenMixin, graphql_jwt.relay.Revoke):
    revoked = graphene.Int(required=False)
    __doc__ = VerifyOrRefreshOrRevokeTokenMixin.__doc__

    class Input:
        refresh_token = graphene.String(required=True)



import time

from django.conf import settings as django_settings
from django.contrib.auth import get_user_model
from django.contrib.sites.shortcuts import get_current_site
from django.core.mail import send_mail
from django.db import models, transaction
from django.template.loader import render_to_string
from django.utils.html import strip_tags

from .constants import TokenAction
from .exceptions import EmailAlreadyInUseError, UserAlreadyVerifiedError, WrongUsageError
from .settings import graphql_auth_settings as app_settings
from .signals import user_verified
from .utils import get_token, get_token_payload

UserModel = get_user_model()


class UserStatus(models.Model):
    """
    A helper model that handles user account stuff.
    """

    user = models.OneToOneField(django_settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name="status")
    verified = models.BooleanField(default=False)
    archived = models.BooleanField(default=False)
    secondary_email = models.EmailField(blank=True, null=True)

    def __str__(self):
        return "%s - status" % (self.user)

    def send(self, subject, template, context, recipient_list=None):
        _subject = render_to_string(subject, context).replace("\n", " ").strip()
        html_message = render_to_string(template, context)
        message = strip_tags(html_message)

        return send_mail(
            subject=_subject,
            from_email=app_settings.EMAIL_FROM,
            message=message,
            html_message=html_message,
            recipient_list=(recipient_list or [getattr(self.user, UserModel.EMAIL_FIELD)]),  # type: ignore
            fail_silently=False,
        )

    def get_email_context(self, info, path, action, **kwargs):
        token = get_token(self.user, action, **kwargs)
        site = get_current_site(info.context)
        return {
            "user": self.user,
            "request": info.context,
            "token": token,
            "port": info.context.get_port(),
            "site_name": site.name,
            "domain": site.domain,
            "protocol": "https" if info.context.is_secure() else "http",
            "path": path,
            "timestamp": time.time(),
            **app_settings.EMAIL_TEMPLATE_VARIABLES,
        }

    def send_activation_email(self, info, *args, **kwargs):
        email_context = self.get_email_context(info, app_settings.ACTIVATION_PATH_ON_EMAIL, TokenAction.ACTIVATION)
        template = app_settings.EMAIL_TEMPLATE_ACTIVATION
        subject = app_settings.EMAIL_SUBJECT_ACTIVATION
        return self.send(subject, template, email_context, *args, **kwargs)

    def resend_activation_email(self, info, *args, **kwargs):
        if self.verified is True:
            raise UserAlreadyVerifiedError
        email_context = self.get_email_context(info, app_settings.ACTIVATION_PATH_ON_EMAIL, TokenAction.ACTIVATION)
        template = app_settings.EMAIL_TEMPLATE_ACTIVATION_RESEND
        subject = app_settings.EMAIL_SUBJECT_ACTIVATION_RESEND
        return self.send(subject, template, email_context, *args, **kwargs)

    def send_password_set_email(self, info, *args, **kwargs):
        email_context = self.get_email_context(info, app_settings.PASSWORD_SET_PATH_ON_EMAIL, TokenAction.PASSWORD_SET)
        template = app_settings.EMAIL_TEMPLATE_PASSWORD_SET
        subject = app_settings.EMAIL_SUBJECT_PASSWORD_SET
        return self.send(subject, template, email_context, *args, **kwargs)

    def send_password_reset_email(self, info, *args, **kwargs):
        email_context = self.get_email_context(
            info, app_settings.PASSWORD_RESET_PATH_ON_EMAIL, TokenAction.PASSWORD_RESET
        )
        template = app_settings.EMAIL_TEMPLATE_PASSWORD_RESET
        subject = app_settings.EMAIL_SUBJECT_PASSWORD_RESET
        return self.send(subject, template, email_context, *args, **kwargs)

    def send_secondary_email_activation(self, info, email):
        if not self.email_is_free(email):
            raise EmailAlreadyInUseError
        email_context = self.get_email_context(
            info,
            app_settings.ACTIVATION_SECONDARY_EMAIL_PATH_ON_EMAIL,
            TokenAction.ACTIVATION_SECONDARY_EMAIL,
            secondary_email=email,
        )
        template = app_settings.EMAIL_TEMPLATE_SECONDARY_EMAIL_ACTIVATION
        subject = app_settings.EMAIL_SUBJECT_SECONDARY_EMAIL_ACTIVATION
        return self.send(subject, template, email_context, recipient_list=[email])

    @classmethod
    def email_is_free(cls, email) -> bool:
        return not UserModel._default_manager.filter(
            models.Q(**{UserModel.EMAIL_FIELD: email}) | models.Q(status__secondary_email=email)  # type: ignore
        ).exists()

    @classmethod
    def clean_email(cls, email=False):
        if email:
            if cls.email_is_free(email) is False:
                raise EmailAlreadyInUseError

    @classmethod
    def verify(cls, token):
        payload = get_token_payload(token, TokenAction.ACTIVATION, app_settings.EXPIRATION_ACTIVATION_TOKEN)
        user = UserModel._default_manager.get(**payload)
        user_status = cls.objects.get(user=user)
        if user_status.verified is False:
            user_status.verified = True
            user_status.save(update_fields=["verified"])
            user_verified.send(sender=cls, user=user)
        else:
            raise UserAlreadyVerifiedError

    @classmethod
    def verify_secondary_email(cls, token):
        payload = get_token_payload(
            token,
            TokenAction.ACTIVATION_SECONDARY_EMAIL,
            app_settings.EXPIRATION_SECONDARY_EMAIL_ACTIVATION_TOKEN,
        )
        secondary_email = payload.pop("secondary_email")
        if not cls.email_is_free(secondary_email):
            raise EmailAlreadyInUseError
        user = UserModel._default_manager.get(**payload)
        user_status = cls.objects.get(user=user)
        user_status.secondary_email = secondary_email
        user_status.save(update_fields=["secondary_email"])

    @classmethod
    def unarchive(cls, user):
        if user.status.archived is True:
            user.status.archived = False
            user.status.save()

    @classmethod
    def archive(cls, user):
        user_status = cls.objects.get(user=user)
        if user_status.archived is False:
            user_status.archived = True
            user_status.save(update_fields=["archived"])

    def swap_emails(self):
        if not self.secondary_email:
            raise WrongUsageError
        with transaction.atomic():
            EMAIL_FIELD = UserModel.EMAIL_FIELD  # type: ignore
            primary = getattr(self.user, EMAIL_FIELD)
            setattr(self.user, EMAIL_FIELD, self.secondary_email)
            self.secondary_email = primary
            self.user.save(update_fields=[EMAIL_FIELD])
            self.save(update_fields=["secondary_email"])

    def remove_secondary_email(self):
        if not self.secondary_email:
            raise WrongUsageError
        with transaction.atomic():
            self.secondary_email = None
            self.save(update_fields=["secondary_email"])



#!/usr/bin/env python

import io
import os
import re
from collections import OrderedDict

from setuptools import find_packages, setup


def get_version(package):
    with io.open(os.path.join(package, "__init__.py")) as f:
        pattern = r'^__version__ = [\'"]([^\'"]*)[\'"]'
        return re.search(pattern, f.read(), re.MULTILINE).group(1)


setup(
    name="django-graphene-auth",
    version=get_version("graphql_auth"),
    license="MIT",
    description="Graphql and relay authentication with Graphene for Django.",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author="ptbang",
    author_email="ptbang@gmail.com",
    maintainer="ptbang",
    url="https://github.com/ptbang/django-graphene-auth",
    project_urls=OrderedDict((("Documentation", "https://django-graphene-auth.readthedocs.io/en/latest/"),)),
    packages=find_packages(exclude=["test*"]),
    install_requires=[
        "Django>=3.2",
        "django-graphql-jwt==0.4.0",
        "django-filter>=23.5",
        "graphene_django>=3.1.6",
        "graphene>=3.3",
        "PyJWT>=2.8.0",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.2",
        "Framework :: Django :: 5.0",
    ],
    keywords="api graphql rest relay graphene auth jwt",
    zip_safe=False,
    include_package_data=True,
)



from graphql_jwt.backends import JSONWebTokenBackend
from graphql_jwt.exceptions import JSONWebTokenError
from graphql_jwt.shortcuts import get_user_by_token
from graphql_jwt.utils import get_credentials


class GraphQLAuthBackend(JSONWebTokenBackend):
    """
    Only difference from the original backend
    is it does not raise when fail on get_user_by_token
    preventing of raise when client send token to a
    mutation that does not re-query login but is not on
    allow any settings.

    Main advantage is to let the mutation handle the
    authentication error.
    """

    def authenticate(self, request=None, **kwargs):
        if request is None or getattr(request, "_jwt_token_auth", False):
            return None

        token = get_credentials(request, **kwargs)

        try:  # +++
            if token is not None:
                return get_user_by_token(token, request)
        except JSONWebTokenError:  # +++
            pass  # +++

        return None



from django import forms
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import UserChangeForm, UserCreationForm, UsernameField

from .settings import graphql_auth_settings as app_settings
from .utils import flat_dict


class RegisterForm(UserCreationForm):
    class Meta:
        model = get_user_model()
        fields = flat_dict(app_settings.REGISTER_MUTATION_FIELDS) + flat_dict(
            app_settings.REGISTER_MUTATION_FIELDS_OPTIONAL
        )


class EmailForm(forms.Form):
    email = forms.EmailField(max_length=254)


class CustomUsernameField(UsernameField):
    required = False


class UpdateAccountForm(UserChangeForm):
    class Meta:
        model = get_user_model()
        fields = flat_dict(app_settings.UPDATE_MUTATION_FIELDS)
        field_classes = {"username": CustomUsernameField}

    def __init__(self, *args, **kwargs):
        super(UpdateAccountForm, self).__init__(*args, **kwargs)
        for key in self.fields.keys():
            self.fields[key].required = False


class PasswordLessRegisterForm(UserCreationForm):
    """
    A RegisterForm with optional password inputs.
    """

    class Meta:
        model = get_user_model()
        fields = flat_dict(app_settings.REGISTER_MUTATION_FIELDS) + flat_dict(
            app_settings.REGISTER_MUTATION_FIELDS_OPTIONAL
        )

    def __init__(self, *args, **kwargs):
        super(PasswordLessRegisterForm, self).__init__(*args, **kwargs)
        self.fields["password1"].required = False
        self.fields["password2"].required = False

    def save(self, commit=True):
        user = super().save(commit=False)
        user.set_unusable_password()
        if commit:
            user.save()
        return user
